/* bfparse.c	brainfuck parser
 *
 * parses a brainfuck program and generates an internal intermediate
 * representation
 *
 * Copyright (C) 2004,2006 Manuel Schiller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "bf.h"

/* amount to grow the memory area holding the intermediate code by */
#define PGM_LENINC	128

/* grow the buffer holding the program's intermediate form */
static void pgm_grow(bfinstr_t **pgm, long *pgmlen)
{
	*pgmlen += PGM_LENINC;
	*pgm = (bfinstr_t *) realloc(*pgm, *pgmlen * sizeof(bfinstr_t));
	assert(*pgm != NULL);
}

/* parse a brainfuck program and write intermediate code */
bfinstr_t *parser(FILE *in)
{
	long loops = 0;	/* make sure that [ and ] match */
	bfinstr_t *pgm = NULL;
	long pgmlen = 0, ip = -1;
	int c;

	pgm_grow(&pgm, &pgmlen);

	while (!feof(in)) {
		c = fgetc(in);
		switch (c) {
		case '+':
			if ((ip == -1) || (pgm[ip]. instr != ADDI)) {
				if (++ip >= pgmlen) pgm_grow(&pgm, &pgmlen);
				pgm[ip].instr = ADDI;
				pgm[ip].attr = 0;
				pgm[ip].attr2 = 0;
			}
			pgm[ip].attr++;
			break;
		case '-':
			if ((ip == -1) || (pgm[ip]. instr != ADDI)) {
				if (++ip >= pgmlen) pgm_grow(&pgm, &pgmlen);
				pgm[ip].instr = ADDI;
				pgm[ip].attr = 0;
				pgm[ip].attr2 = 0;
			}
			pgm[ip].attr--;
			break;
		case '>':
			if ((ip == -1) || (pgm[ip]. instr != PTRI)) {
				if (++ip >= pgmlen) pgm_grow(&pgm, &pgmlen);
				pgm[ip].instr = PTRI;
				pgm[ip].attr = 0;
				pgm[ip].attr2 = 0;
			}
			pgm[ip].attr++;
			break;	
		case '<':
			if ((ip == -1) || (pgm[ip]. instr != PTRI)) {
				if (++ip >= pgmlen) pgm_grow(&pgm, &pgmlen);
				pgm[ip].instr = PTRI;
				pgm[ip].attr = 0;
				pgm[ip].attr2 = 0;
			}
			pgm[ip].attr--;
			break;
		case '[':
			loops++;
			if (++ip >= pgmlen) pgm_grow(&pgm, &pgmlen);
			pgm[ip].instr = LOOP;
			pgm[ip].attr = 0;
			pgm[ip].attr2 = 0;
			break;
		case ']':
			if (--loops < 0) {
				fprintf(stderr, "bfparser.c: Unbalanced ] at " \
					"file offset %ld.\n", ftell(in));
				assert(0 != 0);
			}
			if (++ip >= pgmlen) pgm_grow(&pgm, &pgmlen);
			pgm[ip].instr = POOL;
			pgm[ip].attr = 0;
			pgm[ip].attr2 = 0;
			break;
		case '.':
			if (++ip >= pgmlen) pgm_grow(&pgm, &pgmlen);
			pgm[ip].instr = OUT;
			pgm[ip].attr = 0;
			pgm[ip].attr2 = 0;
			break;
		case ',':
			if (++ip >= pgmlen) pgm_grow(&pgm, &pgmlen);
			pgm[ip].instr = IN;
			pgm[ip].attr = 0;
			pgm[ip].attr2 = 0;
			break;
		default:
			/* anything that is not brainfuck code is a comment */
			break;
		}
	}	
	/* append EOP */
	if (++ip >= pgmlen) pgm_grow(&pgm, &pgmlen);
	pgm[ip].instr = EOP;
	pgm[ip].attr = 0;
	pgm[ip].attr2 = 0;

	/* check for unbalanced [ */
	if (loops > 0) {
		fprintf(stderr, "bfparser.c: Unbalanced [ noticed at " \
			"file offset %ld.\n", ftell(in));
		assert(0 != 0);
	}

	return pgm;
}

/* dump intermediate code to stdout */
void dumper(bfinstr_t *pgm)
{
	long ip = 0;

	do {
		printf("(%5ld)\t", ip++);
		switch (pgm->instr) {
		case ADDI:
			printf("ADDI %ld\n", pgm->attr);
			break;
		case PTRI:
			printf("PTRI %ld\n", pgm->attr);
			break;
		case ADD:
			printf("ADD [%+ld], [cur]*%ld\n", \
			       pgm->attr, pgm->attr2);
			break;
		case ZERO:
			printf("ZERO\n");
			break;
		case LOOP:
			printf("LOOP\n");
			break;
		case POOL:
			printf("POOL\n");
			break;
		case IN:
			printf("IN\n");
			break;
		case OUT:
			printf("OUT\n");
			break;
		case EOP:
			printf("EOP\n");
			break;
		}
	} while((pgm++)->instr != EOP);
}
